package mv3302;

import mv3302.sensor.SimpleConstantRateSensor;
import simkit.SimEntityBase;
import simkit.random.RandomVariate;
import simkit.smd.Mover;

/**
 * The SimpleConstantRateMediator class represents a mediator that manages the
 * interactions between a Mover object and a SimpleConstantRateSensor object. It
 * extends the SimEntityBase class, which provides basic simulation entity
 * functionality.
 */
public final class SimpleConstantRateMediator extends SimEntityBase {

    /**
     * Constructs a SimpleConstantRateMediator object with the specified
     * exponential generator.
     *
     * @param exponentialGenerator the exponential generator to be used by the
     * mediator
     */
    public SimpleConstantRateMediator(RandomVariate exponentialGenerator1) {
        setExponentialGenerator(exponentialGenerator1);
    }

    protected RandomVariate exponentialGenerator;

    /**
     * Constructs a SimpleConstantRateMediator object with the specified
     * exponential generator.
     *
     * @param exponentialGenerator the exponential generator to be used by the
     * mediator
     */
    public void SimpleConstantRateMediator(RandomVariate exponentialGenerator) {
        setExponentialGenerator(exponentialGenerator);
    }

    /**
     * Performs the necessary actions when a Mover object enters the range of a
     * SimpleConstantRateSensor object. If the target is not already detected by
     * the sensor, it waits for a detection delay generated by the exponential
     * generator.
     *
     * @param target the Mover object that entered the sensor's range
     * @param sensor the SimpleConstantRateSensor object that detected the
     * target
     */
    public void doEnterRange(Mover target, SimpleConstantRateSensor sensor) {
        if (!sensor.getContacts().contains(target)) {
            sensor.waitDelay("Detection", exponentialGenerator.generate(), target);
        }
    }

    /**
     * Performs the necessary actions when a Mover object exits the range of a
     * SimpleConstantRateSensor object. If the target is not already undetected
     * by the sensor, it waits for an undetection delay generated by the
     * exponential generator.
     *
     * @param target the Mover object that exited the sensor's range
     * @param sensor the SimpleConstantRateSensor object that lost contact with
     * the target
     */
    public void doExitRange(Mover target, SimpleConstantRateSensor sensor) {
        if (!sensor.getContacts().contains(target)) {
            sensor.waitDelay("Undetection", exponentialGenerator.generate(), target);
        }
    }

    /**
     * Retrieves the exponential generator used by the mediator.
     *
     * @return the exponential generator used by the mediator
     */
    public RandomVariate getExponentialGenerator() {
        return exponentialGenerator;
    }

    /**
     * Sets the exponential generator for the mediator.
     *
     * @param exponentialGenerator the exponential generator to be set
     */
    public void setExponentialGenerator(RandomVariate exponentialGenerator) {
        this.exponentialGenerator = exponentialGenerator;
    }

}
